/*-------------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  temperature-humidity meter by Jack Botner

  main.c - the main program

  This program runs a combined temperature and humidity meter.

  Processor: ATmega164P
      Clock: 8 mHz internal
	Display: 3 digit LCD (simple) LCD-S301C31TR
	Sensor1: TC1046 temperature
	Sensor2: HS1101 humidity

  The humidity sensor is a variable capacitor connected to a
  LMC555 which produces a squarewave output whose frequency is
  a function of the humidity. This is input to the avr's T1 pin
  where the frequency is measured and converted to % hmidity.

  The temperature sensor produces an analog output which is
  applied to the avr's ADC7 pin. The resulting voltage is converted
  to temperature (Celsius).

  The battery voltage is also sampled and applied to the avr's
  ADC0 pin (via 3:1 voltage divider). It is monitored to make sure
  the battery has not run down.
-------------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"

#define TCNT0_INITIAL_VALUE		6	// so timer/counter0 increments 250 times
#define AREF_1P1_MV			1084L	// 1.1V Aref, measured (millivolts)

#define MIN_RH					0
#define MAX_RH				   99

#define STATE_INITIAL			0
#define STATE_ADC0				1
#define STATE_WAIT_2SEC			2
#define STATE_WAIT_HALFSEC		3
#define STATE_ADC7				4
#define STATE_DECIDE			5
#define STATE_TEMPERATURE		6
#define STATE_HUMIDITY			7

#define DIG_SAMPLES		15
#define DIG_ADD			0
#define DIG_GET			1
#define DIG_SORT		2

// Global variables

static volatile uint16_t	uiTenthsBatteryVoltage=0;
static volatile uint16_t	uiLMC555Frequency=0;

// Internal Functions

static int Compare( const void *pElement1, const void *pElement2 );
static int16_t convert_temperature( uint16_t uiAdc );
static uint16_t digital_filter( uint8_t ucRequest, uint16_t uiSample );
static void display_battery(void);
static void display_humidity( uint16_t uiRh );
static void display_temperature( int16_t );
static uint16_t freq2rh( uint16_t uiFreq );
static void system_init(void);

/*-----------------------------------------------------------------
  main()
-----------------------------------------------------------------*/
int main( void )
{
  int8_t 	ii;
  int16_t	iTemp10C;
  uint8_t	ucState=STATE_INITIAL;
  uint16_t	uiAdc, uiTv, uiRh;

  while (1)
  {
    switch ( ucState )
	{
	  case STATE_INITIAL:
		system_init();                   	// initialise system
		ucState = STATE_ADC0;
	    break;

	  case STATE_ADC0:						// sample battery voltage
  		ADMUX  = 0b01000000;				// Vref=AVcc; result right adjusted; select ADC0
  		ADCSRA = 0b10000110;				// Enable ADC; Disable interrupt; prescaler = 64
		set_bit( ADCSRA, ADSC );			// start ADC conversion
		while( !test_bit( ADCSRA, ADIF ) );	// wait until completion
  		uiAdc = (uint16_t) ADCL | ( (uint16_t) ADCH << 8 );
  		uiTv = ( uiAdc * 50 ) / 1024;
  		uiTenthsBatteryVoltage = uiTv * 3;
  		display_battery();
		ucState = STATE_WAIT_2SEC;
	    break;

	  case STATE_WAIT_2SEC:
  		for ( ii=0; ii<60; ++ii )			// 2 sec delay
    	  _delay_loop_2( 65000 );			// 32 ms delay
  		//LcdClear();
		ucState = STATE_ADC7;
		break;

	  case STATE_ADC7:						// sample thermometer
  		ADMUX  = 0b10000111;				// Vref=1.1V; result right adjusted; select ADC7
  		ADCSRA = 0b10000110;				// Enable ADC; Disable interrupt; prescaler = 64
		set_bit( ADCSRA, ADSC );			// start ADC conversion
		while( !test_bit( ADCSRA, ADIF ) );	// wait until completion
  		uiAdc = (uint16_t) ADCL | ( (uint16_t) ADCH << 8 );
		digital_filter( DIG_ADD, uiAdc );
		ucState = STATE_DECIDE;
		break;

	  case STATE_DECIDE:
	    if ( test_bit( PINB, PB2 ) )		// sw2 determines what we see
		  ucState = STATE_TEMPERATURE;
		else
		  ucState = STATE_HUMIDITY;
	    break;

	  case STATE_TEMPERATURE:				// display temperature
		digital_filter( DIG_SORT, 0 );
		uiAdc = digital_filter( DIG_GET, 0 );
		iTemp10C = convert_temperature( uiAdc );
  		display_temperature( iTemp10C );
		ucState = STATE_WAIT_HALFSEC;
	    break;

	  case STATE_HUMIDITY:					// display humidity
		uiRh = freq2rh( uiLMC555Frequency );
  		display_humidity( uiRh );
		ucState = STATE_WAIT_HALFSEC;
	    break;

	  case STATE_WAIT_HALFSEC:
  		for ( ii=0; ii<15; ++ii )			// 2 sec delay
    	  _delay_loop_2( 65000 );			// 32 ms delay
		ucState = STATE_ADC7;
		break;
	}
  }
}

/*-----------------------------------------------------------------
  ISR(TIMER0_OVF_vect)

  Timer 0 overflow interrupt handler 

  Timer 0 is used to generate one-second intervals. Counter 1 is
  read, giving the oscillator frequency in Hz, then reset.

  The clock frequency is 8.0 MHz. Setting the prescaler to 256 
  gives a frequency of 31.250 KHz and a period of 32 us. Now if
  the timer interrupts after 250 counts, giving us an interrupt
  every 8 ms, we can count 125 interrupts to give us exactly
  one second.

  Every second, we initiate an A/D conversion to keep track of
-----------------------------------------------------------------*/
ISR(TIMER0_OVF_vect)
{
  static volatile uint8_t  	ucIntCtr=0;

  ++ucIntCtr;
  if ( ucIntCtr == 125 )		// a second has elapsed
  {
    // Grab timer 1 count
    uiLMC555Frequency = TCNT1;
	TCNT1 = 0;

    // Do an A/D conversion every second
    //clear_bit( PRR, PRADC );		// power up ADC
    set_bit( ADCSRA, ADSC );		// start ADC conversion

    ucIntCtr = 0;					// reset the interrupt counter
  }

  TCNT0 = TCNT0_INITIAL_VALUE;		// starts counter from 6
}

/*-----------------------------------------------------------------
  system_init()

  PORTA: PA0 (input) ADC0 battery voltage sense
		 PA1 (output) LCD "B3" segment
		 PA2 (output) LCD "C3" segment
		 PA3 (output) LCD "D3" segment
		 PA4 (output) LCD "E3" segment
		 PA5 (output) LCD "F3" segment
		 PA6 (output) LCD "G3" segment
		 PA7 (input) ADC7 temperature sensor

  PORTB: PB0 (output) LCD "A3" segment
         PB1 (input) T1 input from humidity sensor
         PB2 (input) SW2 with pull-up
         PB3 (input) Celsius/Fahrenheit jumper with pull-up
         PB4 (output) diagnostic LED
         PB5 (output) MOSI ISP
         PB6 (output) MISO ISP
         PB7 (output) SCK ISP

  PORTC: PC0 (output) LCD "A1" segment
         PC1 (output) LCD "B1" segment
         PC2 (output) LCD "C1" segment
         PC3 (output) LCD "D1" segment
         PC4 (output) LCD "E1" segment
         PC5 (output) LCD "F1" segment
         PC6 (output) LCD "G1" segment
		 PC7 (output) LCD "DP1" segment

  PORTD: PD0 (output) LCD "A2" segment
         PD1 (output) LCD "B2" segment
         PD2 (output) LCD "C2" segment
         PD3 (output) LCD "D2" segment
         PD4 (output) LCD "E2" segment
         PD5 (output) LCD "F2" segment
         PD6 (output) LCD "G2" segment
		 PD7 (output) LCD "DP2" segment
----------------------------------------------------------------*/
static void system_init()
{
  // The fuses are set to provide an internal clock, running at
  // 8.0 MHz with a prescaler of 8. Change the prescaler from
  // 8 to 1. The procedure requires two steps as per the device
  // specifications.
  CLKPR = 0b10000000;       // set clock prescaler change enable
  CLKPR = 0b00000000;       // set division factor to 1

  // Set up the I/O ports directions, pull-ups
  DDRA  = 0b01111110;       // PortA0=in, 1..6=out, 7=input
  PORTA = 0b00000000;       // initialise PortA

  DDRB  = 0b11110001;       // PortB0=output, 1..3=input, 4..7=output
  PORTB = 0b00001100;       // initialise PortB, PB2,3=pullup

  DDRC  = 0b11111111;       // PortC0..7=output
  PORTC = 0b00000000;       // initialise PortC

  DDRD  = 0b11111111;       // PortD0..7 = output
  PORTD = 0b00000000;       // initialise PortD

  // Set up the ADC
  ADMUX  = 0b01000000;		// Vref=AVcc; result right adjusted; select ADC0
  ADCSRA = 0b10000110;		// Enable ADC; Disable interrupt; prescaler = 64
  ADCSRB = 0b00000000;		// Auto trigger source = none
  DIDR0  = 0b10000001;		// Disable ADC0, ADC7 digital input buffers

  // Initialize Timer0. Timer0 is used to create one-second intervals.
  TCCR0A = 0b00000000;      // normal operation, no OC0A, no OC0B
  TCCR0B = 0b00000100;      // prescaler = 256 (turns timer0 on)
  TCNT0 = TCNT0_INITIAL_VALUE;
  TIMSK0 = 0b00000001;      // enable timer/counter0 overflow interrupt

  // Initialize Timer1. Timer1 is used as a counter for the LMC555 oscillator.
  TCCR1A = 0b00000000;		// normal operation, no OC1A, no OC1B
  TCCR1B = 0b00000110;      // external clock on T1 pin, clock on falling edge
  TCCR1C = 0;				// not comparing anything
  TCNT1 = 0;
  TIMSK1 = 0b00000000;      // no timer/counter1 overflow interrupt

  PRR = 0b10010110;			// power save: no TWI, USART1, SPI, USART0
  sei();                    // enable global interrupts
}

/*-----------------------------------------------------------------
  convert_temperature()

  Convert the temperature from the A/D output to 10ths of degrees C.
  ADC0 is monitoring TC1046 Vout signal, and Vref = 1.1V.

  First we need to convert the A/D voltage to millivolts, then
  to tenths of degrees C.
  
  The formula for voltage to temperature conversion is 
  
				  V - 424
			T = ------------
				   6.25
				   
  where T is temperature in degrees C, and V is Vout in millivolts.

  Input: uiAdc 

  Returns: temperature, tenths of degrees Celsius
-----------------------------------------------------------------*/
static int16_t convert_temperature( uint16_t uiAdc )
{
  uint32_t 	ulVoutMv;
  int16_t 	iTemp10C;

  // Input is in the range 0..1023 which represents a voltage of 0..1.1
  // volts (Vref). Scale the adc value to a voltage value (millivolts).

  ulVoutMv = ( (uint32_t) uiAdc * AREF_1P1_MV ) / 1024L;
  // Derive temperature (tens of degrees C) from the formula above.
  // Multiply numerator and denominator by 100 to handle decimals.
  // Multiply numerator by 10 to get tens of degrees.
  
  iTemp10C = (int16_t) ( ( ( (int32_t) ulVoutMv - 424L ) * 1000L ) / 625L );

  if ( iTemp10C > 999 )
    iTemp10C = 999;			// sensor may be disconnected

  return iTemp10C;
}

/*-----------------------------------------------------------------
  display_temperature(...)

  Format the tenths of degrees and display on LCD.
-----------------------------------------------------------------*/
static void display_temperature( int16_t iTemp10C )
{
  int16_t	 	iLen, iTemp;
  char			cBuffer[6], cDisplay[4];
  uint8_t		ucDec;

  if ( iTemp10C > 899 )				// in the oven?
    return;

  if ( test_bit( PINB, PB3 ) )		// jb20111031 was mistakenly PORTB
	iTemp = iTemp10C;						// report degrees C
  else								
	iTemp = ( ( iTemp10C * 9 ) / 5 ) + 320;	// report degrees F

  cDisplay[0] = cDisplay[1] = 0xff;		// in case of error display 0
  cDisplay[2] = '0';
  ucDec = 0;

  itoa( iTemp, cBuffer, 10 );

  // Format the result and insert decimal
  iLen = strlen( cBuffer );

  if ( iLen )
  {
	if ( cBuffer[0] == '-' )
	{
	  // Process negative temperature
	  if ( iLen == 2 )
	  {
	    // Display " -.n"
		cDisplay[0] = 0xff;
		cDisplay[1] = '-';
	    cDisplay[2] = cBuffer[1];
		ucDec = 2;
	  }
	  else
	    if ( iLen == 3 )
		{
		  // Display "-n.n"
		  cDisplay[0] = '-';
		  cDisplay[1] = cBuffer[1];
		  cDisplay[2] = cBuffer[2];
		  ucDec = 2;
		}
		else
		  if ( iLen == 4 )
		  {
		    // Display "-nn" (dropping the tenths digit)
		    cDisplay[0] = '-';
		    cDisplay[1] = cBuffer[1];
		    cDisplay[2] = cBuffer[2];
		    ucDec = 0;
		  }
	}
	else
	{
	  // Process positive temperature
	  ucDec = 2;
	  if ( iLen == 1 )
	  {
	    // Display " 0.n"
		cDisplay[0] = 0xff;
		cDisplay[1] = '0';
	    cDisplay[2] = cBuffer[0];
	  }
	  else
	    if ( iLen == 2 )
		{
		  // Display " n.n"
		  cDisplay[2] = cBuffer[1];
		  cDisplay[1] = cBuffer[0];
		  cDisplay[0] = 0xff;
		}
		else
		{
		  // Display "nn.n"
		  cDisplay[2] = cBuffer[2];
		  cDisplay[1] = cBuffer[1];
		  cDisplay[0] = cBuffer[0];
		}
	}
  }

  LcdOutputNumber( cDisplay[0], 1 );
  LcdOutputNumber( cDisplay[1], 2 );
  LcdOutputNumber( cDisplay[2], 3 );
  LcdOutputDecimal( ucDec );
}

/*-----------------------------------------------------------------
  display_battery()

  Format the tenths of volts and display on LCD.
-----------------------------------------------------------------*/
static void display_battery()
{
  int16_t	 	iLen;
  char			cBuffer[6];

  utoa( uiTenthsBatteryVoltage, cBuffer, 10 );

  // Insert a decimal point in the result
  iLen = strlen( cBuffer );
  if ( !iLen || iLen > 3 )
    return;				// what the heck?

  if ( iLen == 1 )
  {
    cBuffer[2] = cBuffer[0];
	cBuffer[0] = cBuffer[1] = 0xff;
  }
  else
    if ( iLen == 2 )
	{
	  cBuffer[2] = cBuffer[1];
	  cBuffer[1] = cBuffer[0];
	  cBuffer[0] = 0xff;
	}

  LcdOutputNumber( cBuffer[0], 1 );
  LcdOutputNumber( cBuffer[1], 2 );
  LcdOutputNumber( cBuffer[2], 3 );
  LcdOutputDecimal( 2 );
}

/*-----------------------------------------------------------------
  display_humidity(...)

  Format the percent humidity and display on LCD.
-----------------------------------------------------------------*/
static void display_humidity( uint16_t uiRh )
{
  uint16_t	 	uiLen;
  char			cBuffer[6];

  utoa( uiRh, cBuffer, 10 );

  uiLen = strlen( cBuffer );
  if ( !uiLen || uiLen > 2 )
    return;				// what the heck?

  if ( uiLen == 1 )
  {
    cBuffer[2] = cBuffer[0];
	cBuffer[1] = '0';
  }
  else
    if ( uiLen == 2 )
	{
	  cBuffer[2] = cBuffer[1];
	  cBuffer[1] = cBuffer[0];
	}

  cBuffer[0] = 0xff;
  LcdOutputNumber( cBuffer[0], 1 );
  LcdOutputNumber( cBuffer[1], 2 );
  LcdOutputNumber( cBuffer[2], 3 );
  LcdOutputDecimal( 0 );
}

/*-----------------------------------------------------------------
  digital_filter(...)

  ucRequest = DIG_ADD

	Set ucRequest to 0 and pass the sample in uiSample.

  ucRequest = DIG_SORT

	Pass 2 in ucRequest and the array is sorted.

  ucRequest = DIG_GET

	Pass 1 in ucRequest and the median sample is returned.
-----------------------------------------------------------------*/
static uint16_t digital_filter( uint8_t ucRequest, uint16_t uiSample )
{
  static volatile uint16_t uiSamples[DIG_SAMPLES];
  static volatile uint8_t  ucFirst=1;
  uint8_t	ii;
  uint16_t	uiRet;

  if ( ucRequest == DIG_ADD )
  {
	if ( ucFirst )
	{
	  // Initialize the table with first value
	  for ( ii=0; ii<DIG_SAMPLES; ++ii )
	    uiSamples[ii] = uiSample;
	  ucFirst = 0;
	  return 0;
	}

	// A median filtering method is used, in which the new sample
	// is placed in the first and last positions of the array, and
	// the array is sorted in ascending order. Since this code is
	// running from an interrupt handler, schedule the sort later.

    uiSamples[0] = uiSample;
    uiSamples[DIG_SAMPLES-1] = uiSample;
	return 0;
  }

  if ( ucRequest == DIG_GET )
  {
  	// Return the median element
	uiRet = uiSamples[DIG_SAMPLES/2];
	return uiRet;
  }

  if ( ucRequest == DIG_SORT )
	qsort( (void *) &uiSamples, DIG_SAMPLES, sizeof( uint16_t ), Compare );

  return 0;
}

/*--------------------------------------------------------------------------
  Compare(...)
--------------------------------------------------------------------------*/
static int Compare( const void *pElement1, const void *pElement2 )
{
  uint16_t *pui1, *pui2;

  pui1 = (uint16_t *) pElement1;
  pui2 = (uint16_t *) pElement2;

  if ( *pui1 > *pui2 ) return 1;
  if ( *pui1 < *pui2 ) return -1;
  return 0;
}

/*---------------------------------------------------------------------
  freq2rh(...)

  Converts the frequency to relative humidity, using linear interpolation.

  Linear interpolation takes two data points, say (Xa,Ya) and (Xb,Yb),
  and the interpolant (X,Y) is given by:

           (X - Xa) (Yb - Ya)
  Y = Ya + ------------------
               (Xb - Xa)

  In this case X is frequency and Y is relative humidity.

  Input: frequency 7351..6033 Hz
---------------------------------------------------------------------*/
static uint16_t freq2rh( uint16_t uiFreq )
{
  static int16_t sFreq[11] =
  {
    7351, 7224, 7100, 6976, 6853, 6728, 6600, 6468, 6330, 6186, 6033
  };
  static int16_t sRh[11] =
  {
    MIN_RH, 10, 20, 30, 40, 50, 60, 70, 80, 90, MAX_RH
  };

  uint8_t	ii;
  uint16_t	usRH;
  int16_t	sTerm1, sTerm2, sTerm3;

  // Check for boundary values
  if ( uiFreq <= (uint16_t) sFreq[10] ) return MAX_RH;
  if ( uiFreq >= (uint16_t) sFreq[0] )  return MIN_RH;

  for ( ii=0; ii<11; ++ii )
  {
    if ( uiFreq >= (uint16_t) sFreq[ii] )
      break;
  }

  if ( uiFreq == (uint16_t) sFreq[ii] )
    return (uint16_t) sRh[ii];          // no interpolation necessary

  sTerm1 = (int16_t) uiFreq - sFreq[ii-1];
  sTerm2 = sRh[ii] - sRh[ii-1];
  sTerm3 = sFreq[ii] - sFreq[ii-1];
  usRH = (uint16_t) ( sRh[ii-1] + ( ( sTerm1 * sTerm2 ) / sTerm3 ) );

  return usRH;
}
